"""
Plot classes and tools that were once used in Orange widgets

Due to lack of maintenance (non-functioning), the majority of it has been
stripped in this commit
"""

from .owplotgui import *
from .owpalette import *
from .owconstants import *
